from retic import cfg
import gettext

DOMAIN = cfg.APP_LANG_DOMAIN
LOCALEDIR = cfg.APP_LANG_LOCALEDIR
LANGUAGES = cfg.APP_LANG_LANGUAGES
FALLBACK = cfg.APP_LANG_FALLBACK
_s = _t = None

core = gettext.translation(
    "base", "retic/locale",
    languages=["en_US"],
    fallback=True
)
_s = core.gettext

if LOCALEDIR:
    _exists = gettext.find(DOMAIN, LOCALEDIR, LANGUAGES, all=False)
    if not _exists:
        raise FileNotFoundError(_s("ERR_NOT_LOCALE {0}").format(LOCALEDIR))

    t = gettext.translation(
        domain=DOMAIN,
        localedir=LOCALEDIR,
        languages=LANGUAGES,
        fallback=FALLBACK
    )
    _t = t.gettext



from retic.utils.exceptions import get_file_error_exception
from retic.lib.translation.gettext import _s
import traceback
import sys


class HttpMethod(object):
    def __init__(self, method: str, route: any):
        """Initial instance of the HttpMethod Class.

        :param method: HTTP Request Method from the client request
        :param route: Route function from Router class
        """
        self.method = method
        self.route = route

    def default(self, *args):  # self
        """Create a new route in base a path and controllers from the routes config

        :param method: HTTP Request Method from the client request
        :param route: Route function from Router class
        """
        try:
            if len(args) < 2:
                raise Exception(_s('NO_VALID_ROUTE'))
            path = args[0]
            if type(path) is not str:
                raise TypeError(_s('NO_VALID_ROUTE_PATH'))
            self.route(path, self.method.upper(), args[1:])
            return self
        except Exception as e:
            traceback.print_exc(file=sys.stdout)
            print(get_file_error_exception())
            sys.exit(0)



from werkzeug.wrappers import Response
from httpmethods import get_status_by_code
from retic.utils.json import jsonify


class Response(Response):

    def config(
        self,
        environ,
        start_response,
        set_response
    ):
        """Config instance of the Responses Class

        :param environ: Represents a request from a web request.
        :param start_response: Represents a response from a web request.
        :param set_response: Response function for the request."""
        self._environ = environ
        self._start_response = start_response
        self._set_response = set_response
        # set default headers
        self.set_headers('content-type', "application/json")
        return self

    def bad_request(self, content=""):  # Response
        """This method response a *400 Bad Request* response to the client,
        this indicate that the request is invalid.

        This generally means that the request contained invalid parameters
        or headers, or that you tried to do something that your application
        logic does not support."""
        return self._send_by_status(400, content)

    def forbidden(self, content=""):  # Response
        """This method is used to send a *403 Forbidden* response to the client,
        indicating that a request is not authorized.

        This generally means that the user agent tried to do something that they
        were not authorized to do, such as changing another user's password."""
        return self._send_by_status(403, content)

    def not_found(self, content=""):  # Response
        """This method is used to send a *404 Not found* response.

        When called manually from your application code, this method is normally used to
        indicate that the user agent tried to find, update, or delete something that doesn't exist."""
        return self._send_by_status(404, content)

    def ok(self, data: dict = None):  # Response
        """This method is used to send a *200 OK* response to the client."""
        return self._send_by_status(200, jsonify(data) if data else None)

    def server_error(self, content=""):
        """This method is used to send a *500 Server Error* response to the client,
        indicating that some kind of server error occurred."""
        return self._send_by_status(500, content)

    def send(self, data: dict = None):  # Response
        """Send a string response in a non-JSON format (XML, CSV, plain text, etc.).

        This method is used in the underlying implementation of most other terminal response methods."""
        return self.send_string(jsonify(data) if data else {})

    def set_headers(self, headers: dict, value: str = None):  # self
        """Sets the specified response header to the specified value.

        Alternatively, you can pass a single object argument (headers)
        to set multiple header fields at once, where the keys are the names
        of the header fields and the corresponding values ​​are the desired values."""
        if not value and type(headers) is dict:
            self.headers = {**self.headers, **headers}
        else:
            self.headers[headers] = value
        return self

    def set_status(self, code: int):  # self
        """Set the status code for this response."""
        self.status_code = code
        return self

    def send_string(self, data_str: str = ""):  # Response
        """Send a response to http requests"""
        self.set_data(data_str)
        self._set_response(self(self._environ, self._start_response))

    def _send_by_status(self, status, content):
        return self.set_status(status).send_string(content or get_status_by_code(status))

    def favicon(self, req, res):
        return self.ok()



from werkzeug.wrappers import Request
from retic.utils.json import parse


class Request(Request):
    retic = {}
    params = {}
    """Initial instance of the Request Class"""

    def config(self):
        """Set another attributes to Instance"""
        self.body = self._get_body()
        return self

    def param(self, key: str, default_value: str = None):  # str
        """Returns the value of the parameter with the specified name."""
        if key in self.params:
            return self.params.get(key)
        elif key in self.body:
            return self.body.get(key)
        elif key in self.args:
            return self.args.get(key)
        elif key in self.retic:
            return self.retic.get(key)
        return default_value

    def set(self, key: str, value: any = None):  # str
        """Set a value in the requests (req).

        Please note that names are not case sensitive."""
        try:
            return self.retic.setdefault(key.lower(), value)
        except KeyError:
            return None

    def get(self, key: str):  # str
        """Returns the value of the request (req).

        Please note that names are not case sensitive."""
        try:
            return self.retic.get(key.lower(), None)
        except KeyError:
            return None

    def all_params(self):  # dict
        """Returns the value of all the parameters sent in the request,
        combined into a single dictionary.

        It includes parameters parsed from the URL path, the request body,
        and the query string, in that order."""
        return {**self.params, **self.body, **self.args, **self.retic}

    def _get_body(self):
        _content_type = self.headers.get('content-type', type=str)
        try:
            if _content_type.startswith('application/json'):
                _data = parse(self.get_data())
            elif _content_type.startswith('multipart/form-data') \
                    or _content_type.startswith('application/x-www-form-urlencoded'):
                _data = self.form
            else:
                _data = {u"body": self.get_data().decode('utf-8')}
            return _data
        except Exception as e:
            return {}



from retic.lib.hooks.responses import Request, Response, Next
from retic.lib.translation.gettext import _s
from httpmethods import get_http_methods
from inspect import getfile
from .httpmethod import HttpMethod
from .route import Route
from .layer import Layer
import traceback
import sys
import os


class Router(object):
    def __init__(self):
        """Initial instance of the Router Class"""
        self.__dict__ = {
            key: HttpMethod(key, self.route).default for key in get_http_methods()
        }
        self.name = "router"
        self.methods = {key: [] for key in get_http_methods(True)}
        self.args = self.params = []
        self.result = None

    def main(self, environ: dict, start_response: dict):  # []
        """Router main for handle all routes

        :param environ: Request is used to describe an request to a server.
        :param start_response: Represents a response from a web request."""
        try:
            _request = Request(environ).config()
            _response = Response().config(environ, start_response, self._set_response)
            self._endpoint(_request, _response)
            _result = self.result
            self.result = None
            return _result
        except Exception as e:
            return Response(str(e))(environ, start_response)

    def route(self, path: str, method: str, handlers: list):  # Route
        """Create a Route instance and adds this route to specific controller
        of a HTTP method in a Router.

        :param path: Route to searching in the layer
        :param method: HTTP Request Method from the client request
        :param handlers: List of controllers functions from a routes files
        """
        _route = Route(path)
        for _handler in handlers:
            assert _handler, _s('NO_VALID_ROUTE')
            layer = Layer(
                "/",
                {
                    u"strict": False,
                    u"end": True
                },
                _handler
            )
            _route.stack.append(layer)
        _layer = Layer(
            path,
            {
                u"strict": False,
                u"end": True
            },
            _route.dispatch
        )
        _layer.route = _route
        self.methods[method].append(_layer)
        return _route

    def _set_response(self, item: dict):
        """Response function for the request."""
        if self.result:
            print(os.path.abspath(getfile(item)))
            print(_s('ERR_HTTP_HEADERS_SENT'))
            return
        self.result = item

    def _endpoint(self, req: Request, res: Response):  # []
        """This function handle any request from a client, search in the Route List

        :param req: Request is used to describe an request to a server.
        :param res: Represents a response from a web request."""
        try:
            _method = self.methods[req.method]
            if not _method:
                raise KeyError(_s('NO_VALID_METHOD {0}').format(req.method))

            _layer: Layer = self._search_layer(req.path, _method)

            if not _layer:
                raise KeyError(_s('NO_VALID_METHOD {0}').format(req.method))

            _has_method = self._handles_method(_layer)

            assert _has_method, _s('NO_VALID_ROUTE')

            req.params = _layer.params
            return _layer.handle_request(req, res, Next(req, res, _layer))
        except KeyError as e:
            return res.not_found(str(e))
        except Exception as e:
            traceback.print_exc(file=sys.stdout)
            return res.bad_request(str(e))

    def _search_layer(self, path: str, method: str):  # Layer
        """Search a specific Layer using a path and method.

        :param path: Route to searching in the layer
        :param method: HTTP Request Method from the client request
        :return: None if the layer doesn't exists, and a Layer Object
        if the layer is found."""
        _match = None
        for _layer in method:
            _match = _layer.match(path)
            if _match:
                break
        return _match

    def _handles_method(self, layer: Layer):
        """Check if the handle from the Layer is a controller function."""
        return True if layer.route.stack[0].handle else False



from .exceptions import get_file_error_exception
from json import dumps, loads


def jsonify(object):
    """Convert a object to a JSON string."""
    try:
        return dumps(object)
    except Exception as e:
        return dumps({
            u"path": get_file_error_exception(3),
            u"error": str(e)
        })


def parse(str):
    """Deserialize (a str, bytes, or bytearray instance that contains 
    a JSON document) to a Python object.
    """
    return loads(str)



from werkzeug.serving import run_simple
from retic.lib.translation.gettext import _s
from .router import Router

APP_HOST = "0.0.0.0"
APP_PORT = 1801


class App(object):
    def __init__(self):
        self.router = None
        self.apps = {}

    def application(self, environ, start_response):
        """Application for send the response returned by the application to the client

        :param environ: Request is used to describe an request to a server.
        :param start_response: Represents a response from a web request."""
        if not self.router:
            start_response('200 OK', [('Content-Type', 'text/html')])
            return [_s("DEFAULT_RESPONSE").encode("utf8")]
        else:
            return self.router.main(environ, start_response)

    def use(self, item: any, name: str = ""):
        """method of configuring the middleware.

        :param item: item of specific type for specific settings in a app
        :return: instance of the parant type for the item
        """ 

        """TODO: implement another types of item"""
        if isinstance(item, Router):
            self.router = item
        elif name:
            self.apps.setdefault(name, item)

    def listen(
        self,
        hostname=APP_HOST,
        port=APP_PORT,
        application=None,
        use_reloader=False,
        use_debugger=False,
        use_evalex=True,
        extra_files=None,
        reloader_interval=1,
        reloader_type='auto',
        threaded=False,
        processes=1,
        request_handler=None,
        static_files=None,
        passthrough_errors=False,
        ssl_context=None
    ):
        """Create a server based in settings parameters.

        :param hostname: The host to bind to, for example ``'localhost'``.
            If the value is a path that starts with ``unix://`` it will bind
            to a Unix socket instead of a TCP socket..
        :param port: The port for the server.  eg: ``8080``
        :param application: the WSGI application to execute
        :param use_reloader: should the server automatically restart the python
                            process if modules were changed?
        :param use_debugger: should the werkzeug debugging system be used?
        :param use_evalex: should the exception evaluation feature be enabled?
        :param extra_files: a list of files the reloader should watch
                            additionally to the modules.  For example configuration
                            files.
        :param reloader_interval: the interval for the reloader in seconds.
        :param reloader_type: the type of reloader to use.  The default is
                            auto detection.  Valid values are ``'stat'`` and
                            ``'watchdog'``. See :ref:`reloader` for more
                            information.
        :param threaded: should the process handle each request in a separate
                        thread?
        :param processes: if greater than 1 then handle each request in a new process
                        up to this maximum number of concurrent processes.
        :param request_handler: optional parameter that can be used to replace
                                the default one.  You can use this to replace it
                                with a different
                                :class:`~BaseHTTPServer.BaseHTTPRequestHandler`
                                subclass.
        :param static_files: a list or dict of paths for static files.  This works
                            exactly like :class:`SharedDataMiddleware`, it's actually
                            just wrapping the application in that middleware before
                            serving.
        :param passthrough_errors: set this to `True` to disable the error catching.
                                This means that the server will die on errors but
                                it can be useful to hook debuggers in (pdb etc.)
        :param ssl_context: an SSL context for the connection. Either an
                            :class:`ssl.SSLContext`, a tuple in the form
                            ``(cert_file, pkey_file)``, the string ``'adhoc'`` if
                            the server should automatically create one, or ``None``
                            to disable SSL (which is the default).
        source by: werkzeug.serving
        """

        """TODO: Welcome message after the server is created"""
        run_simple(
            hostname=hostname,
            port=int(port),
            application=application or self.application,
            use_reloader=use_reloader,
            use_debugger=use_debugger,
            use_evalex=use_evalex,
            extra_files=extra_files,
            reloader_interval=reloader_interval,
            reloader_type=reloader_type,
            threaded=threaded,
            processes=processes,
            request_handler=request_handler,
            static_files=static_files,
            passthrough_errors=passthrough_errors,
            ssl_context=ssl_context
        )
